// algebra part
#include <gmpxx.h>
#include <sdpa_gmpalgebra.h>

#if defined ENABLE_OPENMP
#include <omp.h>
#endif

#ifdef __cplusplus
extern "C"{
#endif

//http://www.netlib.org/blas/dtrmm.f
//B := alpha*A'*B
int gmp_dtrmmLLTN(int M, int N, mpf_class alpha, mpf_class *A, int LDA, mpf_class *B, int LDB)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
  //#pragma omp for schedule(static)
#endif
  for(int j=0; j<N; j++){
    for(int i=0; i<M; i++){
      mtmp = B[i+j*LDB] *A[i+i*LDA];
      for (int k=i+1; k<M; k++){
	mtmp = mtmp + A[k+i*LDA]*B[k+j*LDB];
      }      
      B[i+j*LDB] = alpha * mtmp;
    }
  }
}

//http://www.netlib.org/blas/dtrmm.f
//B := alpha*A*B
int gmp_dtrmmLUNN(int M, int N, mpf_class alpha, mpf_class *A, int LDA, mpf_class *B, int LDB)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for(int j=0; j<N; j++){
    for(int k=0; k<M; k++){
      mtmp = alpha * B[k+j*LDB];
      for (int i=0; i<k; i++){
	B[i+j*LDB] = B[i+j*LDB]+ mtmp * A[i+k*LDA];
      }
      mtmp = mtmp * A[k+k*LDA];
      B[k+j*LDB] = mtmp;
    }
  }
}


void gmp_dset(int N, mpf_class a, mpf_class* x, int step)
{
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for 
#endif
  for (int i=0; i<N; i++) {
    x[i*step]=a;
  }
}

//http://www.netlib.org/blas/dswap.f
void gmp_dswap(int N, mpf_class *dx, int incx, mpf_class *dy, int incy)
{
  mpf_class mtmp;
  int ix = 0;
  int iy = 0;
  if (incx < 0 ) ix = (-N+1)*incx;
  if (incy < 0 ) iy = (-N+1)*incy;

  for (int i=0; i<N; i++){
    mpf_swap(dx[ix].get_mpf_t(),dy[iy].get_mpf_t());
    ix = ix + incx;
    iy = iy + incy;
  }
}

//http://www.netlib.org/lapack/double/dlanst.f
//DLANST  returns the value of the one norm,  or the Frobenius norm, or
//the  infinity norm,  or the  element of  largest absolute value  of a
//real symmetric tridiagonal matrix A. I means infinite norm.
mpf_class gmp_dlanstI(int N, mpf_class *D, mpf_class *E)
{

  mpf_class anorm;
  anorm = max ( abs(D[0]) + abs(E[0]), abs(E[N-2]) + abs(D[N-1]) );
  for (int i=1; i<=N-2; i++){
    anorm = max ( anorm, abs(D[i]) + abs(E[i]) + abs (E[i-1]) );
  }
  return anorm;
}

int compareGT(mpf_class a, mpf_class b)
{
  if (a>b) return 1;
  else return 0;
}

int compareLT(mpf_class a, mpf_class b)
{
  if (a<b) return 1;
  else return 0;
}

int compare_mpf(const mpf_class *a, const mpf_class *b) 
{
    if (*a>*b)  return 1;
    if (*a==*b) return 0;
    if (*a<*b)  return -1;

}

void gmp_swap(mpf_class *a, mpf_class *b)
{
  mpf_swap(a->get_mpf_t(),b->get_mpf_t());
}

//http://www.netlib.org/blas/daxpy.f
//constant times a vector plus a vector.
int gmp_daxpy(int N, mpf_class alpha, mpf_class *X, int incx, mpf_class *Y, int incy)
{
  int ix = 0;
  int iy = 0;
  for (int i=0; i<N; i++){
    Y[iy] = Y[iy] + alpha * X[ix];
    iy = iy + incy;
    ix = ix + incx;
  }
  return 1;
}

void gmp_clearumat (int N, mpf_class *p, int LDA)
{
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for 
#endif
  for (int i=0; i<N; i++){
    for (int j=i+1; j<N; j++){
      p[i+j*LDA] = 0.0;
    }}
}

void gmp_clearlmat (int N, mpf_class *p, int LDA)
{
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for 
#endif
  for (int i=0; i<N; i++){
    for (int j=0; j<i; j++){
      p[i+j*LDA] = 0.0;
    }}
}

//http://www.netlib.org/blas/dgemm.f
//DGEMM(NN)  performs one of the matrix-matrix operations
//C := alpha*A*B + beta*C,
int gmp_dgemmNN(int M, int N, int K, mpf_class alpha, mpf_class *A, int LDA, mpf_class *B, int LDB, mpf_class beta, mpf_class *C, int LDC)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int i=0;i<M;i++){
    for (int j=0;j<N;j++){
      mtmp = 0.0;
      for (int k=0; k<K; k++){
	mtmp = mtmp + A[i+k*LDA]*B[k+j*LDB];
      }
      C[i+j*LDC] = beta * C[i+j*LDC] + alpha * mtmp;
    }
  }
  return 1;
}

//http://www.netlib.org/blas/dgemm.f
//DGEMM(TN)  performs one of the matrix-matrix operations
//C := alpha*A'*B + beta*C,
int gmp_dgemmTN(int M, int N, int K, mpf_class alpha, mpf_class *A, int LDA, mpf_class *B, int LDB, mpf_class beta, mpf_class *C, int LDC)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int i=0;i<M;i++){
    for (int j=0;j<N;j++){
      mtmp = 0.0;
      for (int k=0; k<K; k++){
	mtmp = mtmp + A[k+i*LDA]*B[k+j*LDB];
      }
      C[i+j*LDC] = beta * C[i+j*LDC] + alpha * mtmp;
    }
  }
  return 1;
}

//http://www.netlib.org/blas/dgemm.f
//DGEMM(NT)  performs one of the matrix-matrix operations
//C := alpha*A*B' + beta*C,
int gmp_dgemmNT(int M, int N, int K, mpf_class alpha, mpf_class *A, int LDA, mpf_class *B, int LDB, mpf_class beta, mpf_class *C, int LDC)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int i=0;i<M;i++){
    for (int j=0;j<N;j++){
      mtmp = 0.0;
      for (int k=0; k<K; k++){
	mtmp = mtmp + A[i+k*LDA]*B[j+k*LDB];
      }
      C[i+j*LDC] = beta * C[i+j*LDC] + alpha * mtmp;
    }
  }
  return 1;
}

//DGEMM(TT)  performs one of the matrix-matrix operations
//C := alpha*A'*B' + beta*C,
int gmp_dgemmTT(int M, int N, int K, mpf_class alpha, mpf_class *A, int LDA, mpf_class *B, int LDB, mpf_class beta, mpf_class *C, int LDC)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int i=0;i<M;i++){
    for (int j=0;j<N;j++){
      mtmp = 0.0;
      for (int k=0; k<K; k++){
	mtmp = mtmp + A[k+i*LDA]*B[j+k*LDB];
      }
      C[i+j*LDC] = beta * C[i+j*LDC] + alpha * mtmp;
    }
  }
  return 1;
}

//uniform random betweek [-1,1]
mpf_class gmp_randomnumber(void)
{
  mpf_class mtmp;
  mtmp = uniformrandom->get_f();
  mtmp = 2.0 * mtmp - 1.0;
  return mtmp;
}

//set random values in a vector
void gmp_setrandvec (int N, mpf_class *p)
{
  for (int i=0; i<N; i++){
    p[i] = gmp_randomnumber();
  }
}

//set random values in a matrix
void gmp_setrandmat (int N, int M, mpf_class *p)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int j=0; j<M; j++){
    for (int i=0; i<N; i++){
      mtmp = gmp_randomnumber();
	p[i+j*N] = mtmp;
    }
  }
}

//set random symmetric matrix.
void gmp_setrandsymmmat (int N, mpf_class *p)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int i=0; i<N; i++){
    for (int j=i; j<N; j++){
      mtmp = gmp_randomnumber();
      p[i+j*N] = mtmp;
      p[j+i*N] = mtmp;
    }
  }
}


// http://www.netlib.org/lapack/double/dlae2.f
//*  Purpose
//*  =======
//*
//*  DLAE2  computes the eigenvalues of a 2-by-2 symmetric matrix
//*     [  A   B  ]
//*     [  B   C  ].
//*  On return, RT1 is the eigenvalue of larger absolute value, and RT2
//*  is the eigenvalue of smaller absolute value.
void gmp_dlae2(mpf_class a, mpf_class b, mpf_class c, mpf_class *rt1, mpf_class *rt2)
{
  mpf_class sm, df, adf, tb, ab;
  mpf_class acmx, acmn, rt;
  mpf_class one, two, half;
  
  one = 1.0;
  two = 2.0;
  half = 0.5;

  sm = a + c;
  df = a - c;
  adf = abs(df);
  tb = b + b;
  ab = abs(tb);

  if (abs(a) > abs(c)) {
    acmx = a;
    acmn = c;
  } else {
    acmx = c;
    acmn = a;
  }

  if (adf > ab) {
    rt = adf * sqrt ( one + ( ab / adf )* ( ab / adf ) );
  } else if (adf < ab ) {
    rt = ab * sqrt ( one + ( adf / ab ) * ( adf / ab ) );
  } else {
    rt = ab * sqrt ( two );
  }
  
  if (sm<0.0) {
    *rt1 = half * (sm-rt);
    *rt2 = ( acmx / (*rt1) ) * acmn - ( b / (*rt1) )*b;
  } else if (sm > 0.0 ) {
    *rt1 = half * ( sm + rt );
    *rt2 = ( acmx / (*rt1) ) * acmn - ( b / (*rt1) )*b;
  } else {
    *rt1 = half*rt;
    *rt2 = -half*rt;
  }
  return;
}

void gmp_setrandlmat (int N, mpf_class *p)
{
  mpf_class mtmp;
  for (int i=0; i<N*N; i++){
    p[i] = 0.0;
  }
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int i=0; i<N; i++){
    for (int j=0; j<=i; j++){
      mtmp = gmp_randomnumber();
      p[i+j*N] = mtmp;
    }
  }
}

void gmp_setrandumat (int N, mpf_class *p)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int i=0; i<N; i++){
    for (int j=i; j<N; j++){
      mtmp = gmp_randomnumber();
      p[i+j*N] = mtmp;
    }
  }
}

//http://www.netlib.org/blas/dtrsm.f
// solves
// X*A'= B, A' is upper triangular matrix (thus A is lower one)
// N indicates that not a unit triangular matrix.
// B is overwritten by X.
// A: lower m times m matrix
// B: m times n matrix
int gmp_dtrsmRLTN(int M, int N, mpf_class alpha, mpf_class *A, int LDA, mpf_class *B, int LDB)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int k=0; k<N; k++){
    mtmp = 1.0 / A[k+k*LDA];
    for (int i=0; i<M; i++){
      B[i+k*LDB] = mtmp * B[i+k*LDB];
    }
    for (int j=k+1; j<N; j++){
      mtmp = A[j+k*LDA];
      for (int i=0; i<M; i++){
	B[i+j*LDB] = B[i+j*LDB] - B[i+k*LDB] * mtmp;
      }
    }
    for(int i=0; i<M; i++){
      B[i+k*LDB] = B[i+k*LDB] * alpha;
    }
  }
  return 1;
}

//http://www.netlib.org/blas/dtrsm.f
// solves
// A'*X= B, A' is lower triangular matrix (thus A is upper one)
// N indicates that not a unit triangular matrix.
// B is overwritten by X.
// A: lower m times m matrix
// B: m times n matrix
int gmp_dtrsmLLTN(int M, int N, mpf_class alpha, mpf_class *A, int LDA, mpf_class *B, int LDB)
{
  mpf_class mtmp;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int j=0; j<N; j++){
    for (int i=M-1; i>=0; i--){
      mtmp = alpha * B[i+j*LDB];
      for (int k=i+1; k<M; k++){
	mtmp = mtmp - A[k+i*LDA]*B[k+j*LDB];
      }
      mtmp = mtmp / A[i+i*LDA];
      B[i+j*LDB] = mtmp;
    }
  }
  return 1;
}

//http://www.netlib.org/blas/dtrsm.f
// solves
// A*X= B, A is lower triangular matrix.
// N indicates that not a unit triangular matrix.
// B is overwritten by X.
// A: lower m times m matrix
// B: m times n matrix
int gmp_dtrsmLLNN(int M, int N, mpf_class alpha, mpf_class *A, int LDA, mpf_class *B, int LDB)
{
  mpf_class zero;
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for 
#endif
  for (int j=0; j<N; j++){
    for (int i=0; i<M; i++){
      B[i+j*LDB] = alpha * B[i+j*LDB];
    }
    for (int k=0; k<M; k++){
      if (B[k+j*LDB]!=zero){
	B[k+j*LDB] = B[k+j*LDB] / A[k+k*LDA];
	for (int i=k+1; i<M; i++){
	  B[i+j*LDB] = B[i+j*LDB] - B[k+j*LDB] * A[i+k*LDA];
	}
      }
    }
  }
}


//http://www.netlib.org/blas/dtrsv.f
//A*x = b
//where A is upper triangular matrix
int gmp_dtrsvUN(int N, mpf_class *A, int LDA, mpf_class *x, int incx)
{
  mpf_class mtmp;
  if(incx!=1) {printf("gmp_dtrsvUN: incx!=1 is not yet supported \n"); exit(-1);}
#if defined ENABLE_OPENMP
  omp_get_num_threads();
#pragma omp parallel for private (mtmp)
#endif
  for (int j=N-1; j>=0; j--){
    if(x[j]!=0.0) {
      x[j] = x[j] / A[j+j*LDA];
      mtmp = x[j];
      for (int i=j-1; i>=0; i--){
	x[i] = x[i] - mtmp * A[i+j*LDA];
      }
    }
  }
  return 1;
}

//http://www.netlib.org/blas/dtrsv.f
//A*x = b
//where A is lower triangular matrix
int gmp_dtrsvLN(int N, mpf_class *A, int LDA, mpf_class *x, int incx)
{
  mpf_class mtmp;
  if(incx!=1) {printf("gmp_dtrsvLN: incx!=1 is not yet supported \n"); exit(-1);}
  for (int j=0; j<N; j++){
    if (x[j] != 0.0 ){
      x[j] = x[j] / A[j+j*LDA];
      mtmp = x[j];
      for (int i=j+1; i<N; i++){
	x[i] = x[i] - mtmp * A[i+j*LDA];
      }
    }
  }
  return 1;
}

//http://www.netlib.org/blas/dtrsv.f
//A'*x = b
//where A is upper triangular matrix
int gmp_dtrsvUT(int N, mpf_class *A, int LDA, mpf_class *x, int incx)
{
  mpf_class mtmp;
  if(incx!=1) {printf("gmp_dtrsvUT: incx!=1 is not yet supported \n"); exit(-1);}
  for (int j=0; j<N; j++){
    mtmp = x[j];
    for (int i=0; i<j; i++){
      mtmp = mtmp - A[i+j*LDA] * x[i];
    }
    mtmp = mtmp / A[j+j*LDA];
    x[j] = mtmp;
  }
  return 1;
}

//http://www.netlib.org/blas/dtrsv.f
//A'*x = b
//where A is lower triangular matrix
int gmp_dtrsvLT(int N, mpf_class *A, int LDA, mpf_class *x, int incx)
{
  mpf_class mtmp;
  if(incx!=1) {printf("gmp_dtrsvLT: incx!=1 is not yet supported \n"); exit(-1);}
  for (int j=N-1; j>=0; j--){
    mtmp = x[j];
    for (int i=N-1; i>j; i--){
      mtmp = mtmp - A[i+j*LDA] * x[i];
    }
    mtmp = mtmp / A[j+j*LDA];
    x[j] = mtmp;
  }
  return 1;
}

//http://www.netlib.org/blas/dtrmv.f
//DTRMV  performs one of the matrix-vector operations
// x := A*x.
// A is upper triangular matrix, N means no-op.
int gmp_dtrmvUN(int N, mpf_class *A, int LDA, mpf_class *x, int incx)
{
  mpf_class mtmp;
  if(incx!=1) {printf("gmp_dtrsvUN: incx!=1 is not yet supported \n"); exit(-1);}

  for (int j=0; j<N; j++){
    if ( x[j] != 0.0 ) {
      mtmp = x[j];
      for (int i=0; i<=j-1; i++){
	x[i] = x[i] + mtmp * A[i+j*LDA];
      }
      x[j] = x[j] * A [j+j*LDA];
    }
  }
  return 1;
}

//http://www.netlib.org/blas/dtrmv.f
//DTRMV  performs one of the matrix-vector operations
// x := A'*x.
// A is upper triangular matrix, T means taking transpose of A.
int gmp_dtrmvUT(int N, mpf_class *A, int LDA, mpf_class *x, int incx)
{
  mpf_class mtmp;
  if(incx!=1) {printf("gmp_dtrsvUT: incx!=1 is not yet supported \n"); exit(-1);}

  for (int j=N-1; j>=0; j--){
    mtmp = x[j] * A[j+j*LDA];
    for (int i=j-1; i>=0; i--){
      mtmp = mtmp +  A[i+j*LDA] * x[i];
    }
    x[j] = mtmp;
  }
  return 1;
}

//http://www.netlib.org/blas/dtrmv.f
//DTRMV  performs one of the matrix-vector operations
// x := A*x.
// A is lower triangular matrix, T means taking transpose of A.
int gmp_dtrmvLN(int N, mpf_class *A, int LDA, mpf_class *x, int incx)
{
  mpf_class mtmp;
  if(incx!=1) {printf("gmp_dtrsvLN: incx!=1 is not yet supported \n"); exit(-1);}

  for (int j=N-1; j>=0; j--){
    mtmp = x[j];
    for (int i=N-1; i>=j+1; i--){
      x[i] = x[i] + mtmp * A[i+j*LDA];
    }
    x[j] = x[j] * A[j+j*LDA];
  }
  return 1;
}

//http://www.netlib.org/blas/dtrmv.f
//DTRMV  performs one of the matrix-vector operations
// x := A'*x.
// A is lower triangular matrix, T means taking transpose of A.
int gmp_dtrmvLT(int N, mpf_class *A, int LDA, mpf_class *x, int incx)
{
  mpf_class mtmp;
  if(incx!=1) {printf("gmp_dtrsvLT: incx!=1 is not yet supported \n"); exit(-1);}

  for (int j=0; j<N; j++){
    mtmp = x[j];
    mtmp = mtmp * A[j+j*LDA];
    for (int i=j+1; i<N; i++){
      mtmp = mtmp + A[i+j*LDA] * x[i];
    }
    x[j] = mtmp;
  }
  return 1;
}

//http://www.netlib.org/blas/dcopy.f
//copies a vector, x, to a vector, y.
void gmp_dcopy(int N, mpf_class *X, int incx, mpf_class *Y, int incy)
{
  if(N<=0) return;
  if( incx<0 || incy<0 ) {printf("gmp_dcopy: incx<0 or incy<0 case are not yet supported \n"); exit(-1);}
  int ix = 0, iy = 0;
  for(int i=0;i<N; i++){
    Y[iy]=X[ix];
    ix = ix + incx;
    iy = iy + incy;
  }
  return;
}

//http://www.netlib.org/blas/dgemv.f
//DGEMV  performs one of the matrix-vector operations
//y := alpha*A*x + beta*y, N means no-op for A.
//A: m times n matrix
//x: n dimension
//y: m dimension
int gmp_dgemvN(int M, int N, mpf_class alpha, mpf_class *A, int LDA, mpf_class *x, int incx, mpf_class beta, mpf_class *y, int incy)
{
  mpf_class mtmp;
  if(incx!=1 ||incy!=1 ) {printf("gmp_dgemvN: incx!=1 or incy!=1 case are not yet supported \n"); exit(-1);}
  //First form  y := beta*y.
  if(beta == 0.0 ){
    for (int i=0; i<M; i++){
      y[i] = 0.0;
    }
  } else {
    for (int i=0; i<M; i++){
      y[i] = beta * y[i];
    }
  }
  //Form  y := alpha*A*x + y.
  for (int j=0; j<N; j++){
    mtmp = alpha * x[j];
    for (int i=0; i<M; i++){
      y[i] = y[i] + mtmp * A[i+j*LDA];
    }
  }
}
  
//http://www.netlib.org/blas/dgemv.f
//DGEMV  performs one of the matrix-vector operations
//y := alpha*A'*x + beta*y, (T means transpose)
//A: m times n matrix
//x: m dimension
//y: n dimension
int gmp_dgemvT(int M, int N, mpf_class alpha, mpf_class *A, int LDA, mpf_class *x, int incx, mpf_class beta, mpf_class *y, int incy)
{
  mpf_class mtmp;
  if(incx!=1 ||incy!=1 ) {printf("gmp_dgemvT: incx!=1 or incy!=1 case are not yet supported \n"); exit(-1);}
  //First form  y := beta*y.
  if(beta == 0.0 ){
    for (int i=0; i<N; i++){
      y[i] = 0.0;
    }
  } else {
    for (int i=0; i<N; i++){
      y[i] = beta * y[i];
    }
  }
  //Form  y := alpha*A'*x + y.
  for (int j=0; j<M; j++){
    mtmp = alpha * x[j];
    for (int i=0; i<N; i++){
      y[i] = y[i] + mtmp * A[j+i*LDA];
    }
  }
}
	
//http://www.netlib.org/blas/ddot.f
//forms the dot product of two vectors.
//returns a'*b
mpf_class gmp_ddot(int n, mpf_class *dx, int incx, mpf_class *dy, int incy)
{
  mpf_class ddot;
  mpf_class dtmp;
  int ix, iy;
  ddot = 0.0;
  dtmp = 0.0;
  if (n<=0) return ddot;
  if (incx!=1 || incy!=1) {
//code for unequal increments or equal increments
//not equal to 1
    ix = 0;
    iy = 0;
    if(incx<0) ix = (-n+1)*incx + 1;
    if(incy<0) iy = (-n+1)*incy + 1;
    for (int i=0; i<n; i++){
        dtmp = dtmp + dx[ix]*dy[iy];
        ix = ix + incx;
        iy = iy + incy;
    }
    ddot = dtmp;
    return ddot;
  } else {
//code for both increments equal to 1
    for(int i=0; i<n; i++){
      dtmp=dtmp+dx[i]*dy[i];
    }
  return dtmp;
  }
}

//reduces a real symmetric matrix A to real symmetric
//tridiagonal form T by an orthogonal similarity transformation:
//see also: http://www.netlib.org/lapack/double/dsytrd.f
//this function accesses all elements, not upper or lower part only.
int gmp_dsytrd_old(int N, mpf_class *A, int LDA, mpf_class *D, mpf_class *E){
  mpf_class h, hh, k;
  mpf_class mtmp;
  int p;
  for(int m=N-1; m>=1; m--) {
  //calculate sigma
      h = gmp_ddot(m, A+m, LDA, A+m, LDA);
  //only we need a_{i,i-1} + sigma (h)
    hh=sqrt(h);
    if (A[m+(m-1)*LDA] < 0.0){
      hh=-hh;
    }
    //calculate H = 1/2 |u|^2, and now u = (A[m,0],...,A[m,m-1])
    h = h + A[m+(m-1)*LDA] * hh;
    A[m+(m-1)*LDA] = A[m+(m-1)*LDA] + hh;
    //calculate p in E.
      for (int r=0; r<m; r++) {
	mtmp=0.0;
	for (int s=0; s<m; s++) {
	  mtmp=mtmp+A[r+(s*LDA)]*A[m+(s*LDA)];
	}
	E[r]=mtmp/h;
      }
    //calculate K
      k = gmp_ddot (m, E, 1, A+m, LDA) / ( 2.0 * h );
    //calculate q
      for (int r=0; r<m; r++) {
	E[r] = E[r] - k * A[m+r*LDA];
      }
    //calculate A'
     for(int i=0; i<m; i++){
       for(int j=0; j<m; j++){
	 A[i+j*LDA] = A[i+j*LDA] - E[i]*A[m+j*LDA] - A[m+i*LDA]*E[j];
       }
     }
  //Diagonal part
    D[m]=A[m+m*LDA];
  //super diagonal part
    E[m-1]=hh;
  }
  D[0]=A[0];
  return 1;
}



//http://www.netlib.org/lapack/double/dpotf2.f
//see also http://www.netlib.org/lapack/double/dpotrf.f
//the Cholesky factorization of a real symmetric positive definite matrix A.
int gmp_dpotrfU(int N, mpf_class *A, int LDA)
{
  mpf_class ftmp;
  for (int i=0; i<N; i++){
  // r_{ii}
    ftmp=0.0;
    for (int j=0; j<i; j++){
      ftmp=ftmp + A[j+i*LDA] * A[j+i*LDA];
    }
    if( (A[i+i*LDA]-ftmp) < 0.0 ) return i+1;
    A[i+i*LDA] = sqrt(A[i+i*LDA] - ftmp);
  // r_{ij}
    for (int j=i+1; j<N; j++){
      ftmp=0.0;
      for (int k=0; k<i; k++){
        ftmp = ftmp + A[k+i*LDA] * A[k+j*LDA];
      }
      A[i+j*LDA] = (A[i+j*LDA] - ftmp) / A[i+i*LDA];
    }
  }
  return 0; //successful exit
}

//http://www.netlib.org/lapack/double/dpotf2.f
//see also http://www.netlib.org/lapack/double/dpotrf.f
//the Cholesky factorization of a real symmetric positive definite matrix A.
int gmp_dpotrfL(int N, mpf_class *A, int LDA)
{
 mpf_class ftmp;
  for (int i=0; i<N; i++){
  // l_{ii}
    ftmp=0.0;
    for (int j=0; j<i; j++){
      ftmp=ftmp + A[i+j*LDA] * A[i+j*LDA];
    }
  if( (A[i+i*LDA]-ftmp) < 0.0 ) { return i+1;}
    A[i+i*LDA] = sqrt(A[i+i*LDA] - ftmp);
  // l_{ij}
    for (int j=i+1; j<N; j++){
      ftmp=0.0;
      for (int k=0; k<=i-1; k++){
        ftmp = ftmp + A[j+k*LDA] * A[i+k*LDA];
      }
      A[j+i*LDA] = (A[j+i*LDA] - ftmp) / A[i+i*LDA];
    }
  }
  return 0; //successful exit
}

//http://www.netlib.org/lapack/double/dpotf2.f
//see also http://www.netlib.org/lapack/double/dpotrf.f
//the Cholesky factorization of a real symmetric positive definite matrix A.
//By YAMASHITA Makoto
int gmp_dpotrfLi(int N, mpf_class* A, int LDA)
{
  mpf_class ftmp;
  for (int i=0; i<N; ++i) {
    ftmp = A[i + i*LDA];
    if (ftmp < 0.0) {
      return (i+1);
    }
    ftmp = sqrt(ftmp);
    A[i + i*LDA] = ftmp;
    ftmp = 1.0/ftmp;
    for (int j=i+1; j<N; ++j) {
      A[j + i*LDA] = A[j + i*LDA]*ftmp;
    }
    for (int j1=i+1; j1<N; ++j1) {
      for (int i1=j1; i1<N; ++i1) {
        A[i1 + j1*LDA] = A[i1 + j1*LDA] - A[i1 + i*LDA]*A[j1 + i*LDA];
      }
    }
  }
  return 0;
}

//http://www.netlib.org/blas/dnrm2.f
//DNRM2 returns the euclidean norm of a vector via the function
//so that DNRM2 := sqrt( x'*x ).
mpf_class gmp_dnrm2(int N, mpf_class *x, int inc)
{
  mpf_class zero = 0.0;
  mpf_class one  = 1.0;
  mpf_class mtmp, ssq, scale, absxi;
  if (N<1)  return zero;
  if (N==1) {
    mtmp = abs( x[0] );
    return mtmp;
  }
  scale = zero;
  ssq = one;
  for (int i=0; i<N*inc; i=i+inc){
    if (x[i]!=zero){
      absxi = abs( x[i] );
      if (scale < absxi ) {
        ssq = one + ssq * (scale/absxi) * (scale/absxi);
        scale = absxi;
      } else {
        ssq = ssq + ( absxi/scale ) * ( absxi/scale );
      }
    }
  }
  mtmp = scale * sqrt (ssq);
  return mtmp;
}

//http://www.netlib.org/blas/dscal.f
//scales a vector by a constant.
int gmp_dscal(int N, mpf_class alpha, mpf_class *X, int inc)
{
  for (int i=0; i<N*inc; i=i+inc){
    X[i]=alpha*X[i];
  }
  return 1;
}

//http://www.netlib.org/lapack/double/dlapy2.f
//DLAPY2 returns sqrt(x**2+y**2), taking care not to cause unnecessary
//overflow.
mpf_class gmp_dlapy2(mpf_class x, mpf_class y)
{
  mpf_class zero = 0.0;
  mpf_class one = 1.0;
  mpf_class w, z;

  mpf_class xabs, yabs;
  xabs = abs (x);
  yabs = abs (y);
  w = max (xabs, yabs);
  z = min (xabs, yabs);
  if (z == 0.0 ){
    return w;
  } else {
    w = w * sqrt ( one + (z/w) * (z/w) );
    return w;
  }
//not reached
}

//cf.http://www.netlib.org/blas/dlamch.f
//E denots we always calculate relative machine precision (e).
//where 1+e > 1, minimum of e.
//XXX implementation is not reliable.
mpf_class gmp_dlamchE(void) {
  mpf_class eps;
  mpf_class o_eps;
  mpf_class one;
  eps = 1.0;
  one = 1.0;
  while (one+eps > one) {
    o_eps = eps;
    eps=eps/2.0;
  }
  return o_eps;
}

//DLAMCH determines machine parameters.
//S denots we always calculate `safe minimum, such that 1/sfmin does not overflow'.
//cf.http://www.netlib.org/blas/dlamch.f
mpf_class gmp_dlamchS(void)
{
  int exp;
  mpf_class largeval, sfmin;
  largeval=2.0;
  exp=32;
//  if(sizeof(void*)==4) exp = 32;
//  if(sizeof(void*)==8) exp = 64;
  if(exp==0) { printf("gmp_dlamchS : XXX do not support this arch\n"); }

  for(int i=0; i<exp; i++)
    largeval=largeval*largeval;
  sfmin = 1.0 / largeval;
  return sfmin;
}

//DLAMCH determines machine parameters.
//P   prec = eps*base
//cf.http://www.netlib.org/blas/dlamch.f
mpf_class gmp_dlamchP(void)
{
  mpf_class base, eps, prec;

  base = gmp_dlamchB();
  eps = gmp_dlamchE();
  prec = eps * base;
  return prec;
}

//DLAMCH determines machine parameters.
//B  base  = base of the machine
//cf.http://www.netlib.org/blas/dlamch.f
mpf_class gmp_dlamchB(void)
{
  mpf_class two;
  two = 2.0;
  return two;
}

//implementation of sign transfer function.
mpf_class gmp_sign(mpf_class a, mpf_class b)
{
  mpf_class mtmp;
  if (b>=0.0){
    mtmp=abs(a);
  } else {
    mtmp=-abs(a);
  }
  return mtmp;
}

//http://www.netlib.org/lapack/double/dlarfg.f
// DLARFG generates a real elementary reflector H of order n, such
// that
//       H * ( alpha ) = ( beta ),   H' * H = I.
//           (   x   )   (   0  )
// where alpha and beta are scalars, and x is an (n-1)-element real vector.
void gmp_dlarfg(int N, mpf_class *alpha, mpf_class *x, int incx, mpf_class *tau)
{
  mpf_class xnorm;
  mpf_class zero;
  mpf_class one;
  mpf_class beta;
  mpf_class safmin;
  mpf_class rsafmn;
  int knt;

  zero = 0.0;
  one = 1.0;

  if (N<=1) {
    *tau = 0.0;
    return;
  }
  xnorm = gmp_dnrm2(N-1, x, incx);
//H  =  I
  if (xnorm == 0.0 ) {
    *tau = 0.0;
  } else {
    beta = -1.0 * gmp_sign( gmp_dlapy2( *alpha, xnorm ), *alpha );
    safmin = gmp_dlamchS()/gmp_dlamchE();
//XNORM, BETA may be inaccurate; scale X and recompute them
    if (abs(beta)<safmin) {
      printf("# gmp_dlarfg: XXX not very well tested\n");
// gmp_printf("# gmp_dlarfg: |beta|<safmin "P_FORMAT" \n", safmin.get_mpf_t() );
      rsafmn = one / safmin;
      knt = 0;
      while (abs(beta)<safmin) {
        knt++;
        gmp_dscal(N-1, rsafmn, x, incx);
        beta = beta * rsafmn;
        *alpha = *alpha * rsafmn;
      }
//New BETA is at most 1, at least SAFMIN
      xnorm = gmp_dnrm2(N-1, x, incx);
      beta = - 1.0 * gmp_sign(gmp_dlapy2 ( *alpha, xnorm ), *alpha);
      *tau = (beta - *alpha ) /beta;
      gmp_dscal( N-1, one / ( *alpha-beta), x, incx );
//If ALPHA is subnormal, it may lose relative accuracy
      *alpha = beta;
      for (int j=0; j<knt; j++){
        *alpha = *alpha * safmin;
      }
    } else {
      *tau = (beta-*alpha)/beta;
      gmp_dscal( N-1, one / ( *alpha-beta), x, incx );
      *alpha = beta;
    }
  }
}

//http://www.netlib.org/blas/dsyr2.f
//DSYR2  performs the symmetric rank 2 operation
//   A := alpha*x*y' + alpha*y*x' + A,
//where alpha is a scalar, x and y are n element vectors and A is an n
//by n symmetric matrix. U means upper triangle of A is referenced.
void gmp_dsyr2U(int N, mpf_class alpha, mpf_class *x, int incx, mpf_class *y, int incy, mpf_class *A, int LDA)
{
  mpf_class mtmp1, mtmp2;
  int ix, iy, jx, jy, kx, ky;

  if (N==0) return;
  if (alpha==0.0) return;
//    Set up the start points in X and Y if the increments are not both
//    unity.
  if(incx!=1 || incy!=1 ) {
    if(incx>0) kx = 0;
  else {kx = 1 - ( N - 1 )*incx; printf("# gmp_dsyr2U incx<0 case is not tested\n"); exit(-1);}
    if(incy>0) ky = 0;
  else {ky = 1 - ( N - 1 )*incy; printf("# gmp_dsyr2U incy<0 case is not tested\n"); exit(-1);}
    jx=kx;
    jy=ky;
  }
  if (incx==1 && incy==1) {
    for (int j=0; j<N; j++){
      mtmp1 = alpha*y[j];
      mtmp2 = alpha*x[j];
      for (int i=0; i<=j; i++){
        A[i+j*LDA] = A[i+j*LDA] + x[i]*mtmp1 + y[i]*mtmp2;
      }
    }
  } else {
    for (int j=0; j<N; j++){
      mtmp1 = alpha*y[jy];
      mtmp2 = alpha*x[jx];
      ix = kx;
      iy = ky;
      for (int i=0; i<=j; i++){
        A[i+j*LDA] = A[i+j*LDA] + x[ix]*mtmp1 + y[iy]*mtmp2;
        ix = ix + incx;
        iy = iy + incy;
      }
      jx = jx + incx;
      jy = jy + incy;
    }
  }
}
//http://www.netlib.org/blas/dsyr2.f
//DSYR2  performs the symmetric rank 2 operation
//   A := alpha*x*y' + alpha*y*x' + A,
//where alpha is a scalar, x and y are n element vectors and A is an n
//by n symmetric matrix. L means lower triangle of A is referenced.
void gmp_dsyr2L(int N, mpf_class alpha, mpf_class *x, int incx, mpf_class *y, int incy, mpf_class *A, int LDA)
{
  mpf_class mtmp1, mtmp2;
  int ix, iy, jx, jy, kx, ky;

  if (N==0) return;
  if (alpha==0.0) return;
//    Set up the start points in X and Y if the increments are not both
//    unity.
  if(incx!=1 || incy!=1 ) {
    if(incx>0) kx = 0;
  else {kx = 1 - ( N - 1 )*incx; printf("# gmp_dsyr2L incx<0 case is not tested\n"); exit(-1);}
    if(incy>0) ky = 0;
  else {ky = 1 - ( N - 1 )*incy; printf("# gmp_dsyr2L incy<0 case is not tested\n"); exit(-1);}
    jx=kx;
    jy=ky;
  }
  if (incx==1 && incy==1) {
    for (int j=0; j<N; j++){
      mtmp1 = alpha*y[j];
      mtmp2 = alpha*x[j];
      for (int i=0; i<=j; i++){
        A[j+i*LDA] = A[j+i*LDA] + x[i]*mtmp1 + y[i]*mtmp2;
      }
    }
  } else {
    for (int j=0; j<N; j++){
      mtmp1 = alpha*y[jy];
      mtmp2 = alpha*x[jx];
      ix = kx;
      iy = ky;
      for (int i=0; i<=j; i++){
        A[j+i*LDA] = A[j+i*LDA] + x[ix]*mtmp1 + y[iy]*mtmp2;
        ix = ix + incx;
        iy = iy + incy;
      }
      jx = jx + incx;
      jy = jy + incy;
    }
  }
}

//http://www.netlib.org/blas/dsymv.f
//DSYMV  performs the matrix-vector  operation
//y := alpha*A*x + beta*y,
//where alpha and beta are scalars, x and y are n element vectors and
//A is an n by n symmetric matrix.
void gmp_dsymvL(int N, mpf_class alpha, mpf_class *A, int LDA, mpf_class *x, int incx, mpf_class beta, mpf_class *y, int incy)
{

  int ix, iy, jx, jy, kx, ky;
  mpf_class mtmp1, mtmp2;
  mpf_class zero;
  mpf_class one;

  zero = 0.0;
  one = 1.0;

  if (N==0 || ((alpha==zero) && (beta==one)) ) return;
//    Set up the start points in  X  and  Y.
  if(incx>0) kx = 0;
  else kx = - ( N - 1 )*incx;
  if(incy>0) ky = 0;
  else ky = - ( N - 1 )*incy;
//    Start the operations. In this version the elements of A are
//    accessed sequentially with one pass through the triangular part
//    of A.
  //    First form  y := beta*y.
  if(beta!=one) {
    if(incy==1) {
      for(int i=0; i<N; i++){
        y[i] = beta*y[i];
      }
    } else {
      for(int i=0; i<N; i++){
        y[iy]=beta*y[iy];
        iy = iy + incy;
      }
    }
  }
//Form  y  when A is stored in upper triangle.
  if(incx==1 && incy==1){
    for(int j=0; j<N; j++) {
      mtmp1=alpha*x[j];
      mtmp2=zero;
      for(int i=0; i<j; i++){
        y[i] = y[i]+ mtmp1 * A[j+i*LDA];
        mtmp2 = mtmp2 + A[j+i*LDA]*x[i];
      }
      y[j] = y[j]+ mtmp1 * A[j+j*LDA]+ alpha * mtmp2;
    }
  } else {
    jx = kx;
    jy = ky;
    for(int j=0; j<N; j++){
      mtmp1 = alpha*x[jx];
      mtmp2 = zero;
      ix = kx;
      iy = ky;
      for (int i=0; i<j; i++){
        y[iy]=y[iy] + mtmp1 * A[j+i*LDA];
        mtmp2 = mtmp2 + A[j+i*LDA]*x[ix];
        ix = ix + incx;
        iy = iy + incy;
      }
      y[jy] = y[jy] + mtmp1 * A[j+j*LDA] + alpha * mtmp2;
      jx = jx + incx;
      jy = jy + incy;
    }
  }
}

//http://www.netlib.org/blas/dsymv.f
//DSYMV  performs the matrix-vector  operation
//   y := alpha*A*x + beta*y,
//where alpha and beta are scalars, x and y are n element vectors and
//A is an n by n symmetric matrix.
void gmp_dsymvU(int N, mpf_class alpha, mpf_class *A, int LDA, mpf_class *x, int incx, mpf_class beta, mpf_class *y, int incy)
{

  int ix, iy, jx, jy, kx, ky;
  mpf_class mtmp1, mtmp2;
  mpf_class zero;
  mpf_class one;

  zero = 0.0;
  one = 1.0;

  if (N==0 || ((alpha==zero) && (beta==one)) ) return;
//    Set up the start points in  X  and  Y.
  if(incx>0) kx = 0;
  else kx = - ( N - 1 )*incx;
  if(incy>0) ky = 0;
  else ky = - ( N - 1 )*incy;
//    Start the operations. In this version the elements of A are
//    accessed sequentially with one pass through the triangular part
//    of A.
  //    First form  y := beta*y.
  if(beta!=one) {
    if(incy==1) {
      for(int i=0; i<N; i++){
        y[i] = beta*y[i];
      }
    } else {
      for(int i=0; i<N; i++){
        y[iy]=beta*y[iy];
        iy = iy + incy;
      }
    }
  }
//Form  y  when A is stored in upper triangle.
  if(incx==1 && incy==1){
    for(int j=0; j<N; j++) {
      mtmp1=alpha*x[j];
      mtmp2=zero;
      for(int i=0; i<j; i++){

        y[i] = y[i]+ mtmp1 * A[i+j*LDA];
        mtmp2 = mtmp2 + A[i+j*LDA]*x[i];
      }
      y[j] = y[j]+ mtmp1 * A[j+j*LDA]+ alpha * mtmp2;
    }
  } else {
    jx = kx;
    jy = ky;
    for(int j=0; j<N; j++){
      mtmp1 = alpha*x[jx];
      mtmp2 = zero;
      ix = kx;
      iy = ky;
      for (int i=0; i<j; i++){
        y[iy]=y[iy] + mtmp1 * A[i+j*LDA];
        mtmp2 = mtmp2 + A[i+j*LDA]*x[ix];
        ix = ix + incx;
        iy = iy + incy;
      }
      y[jy] = y[jy] + mtmp1 * A[j+j*LDA] + alpha * mtmp2;
      jx = jx + incx;
      jy = jy + incy;
    }
  }
}

//http://www.netlib.org/lapack/double/dsytd2.f
//see also: http://www.netlib.org/lapack/double/dsytrd.f
// reduces a real symmetric matrix A to real symmetric
// tridiagonal form T by an orthogonal similarity transformation:
// Q**T * A * Q = T.
// N is the dimension of matrix
// A is NxN matrix, overwritten
// D is N   array; diagonal part
// E is N-1 array; superdiagonal part
void gmp_dsytrdL(int N, mpf_class *A, int LDA, mpf_class *d, mpf_class *e, mpf_class *tau)
{
  mpf_class one;
  mpf_class mone;
  mpf_class zero;
  mpf_class half;
  mpf_class taui;
  mpf_class alpha;
  one = 1.0;
  mone = -1.0;
  zero = 0.0;
  half = 1.0/2.0;

  if(N<=0) return;
// Reduce the lower triangle of A
  for (int i=0; i<N-1; i++){
//Generate elementary reflector H(i) = I - tau * v * v'
    //to annihilate A(i+2:n,i)
    gmp_dlarfg( N-i-1, &A[(i+1)+i*LDA], &A[(min(i+2,N))+i*LDA], 1, &taui);
    e[i] = A[(i+1)+i*LDA];
    if (taui!=zero){
      //Apply H(i) from both sides to A(i+1:n,i+1:n)
      A[(i+1)+i*LDA] = one;
      //Compute  x := tau * A * v  storing y in TAU(i:n-1)
      gmp_dsymvL(N-i-1, taui, &A[(i+1)+(i+1)*LDA], N, &A[(i+1)+i*LDA], 1, zero, &tau[i], 1);
      //Compute  w := x - 1/2 * tau * (x'*v) * v
      alpha = mone * half * taui * gmp_ddot( N-i-1, &tau[i], 1, &A[(i+1)+i*LDA], 1 );
      gmp_daxpy( N-i-1, alpha, &A[(i+1)+i*LDA], 1, &tau[i], 1 );
      //Apply the transformation as a rank-2 update:
      //A := A - v * w' - w * v'
      gmp_dsyr2L( N-i-1, mone, &A[(i+1)+i*LDA], 1, &tau[i], 1, &A[(i+1)+(i+1)*LDA], N);
      A[(i+1)+i*LDA] = e[i];
    }
    d[i] = A[i+i*LDA];
    tau[i]=taui;
  }
  d[N-1] = A[N*N-1];
}

//http://www.netlib.org/lapack/double/dsytd2.f
//see also: http://www.netlib.org/lapack/double/dsytrd.f
// reduces a real symmetric matrix A to real symmetric
// tridiagonal form T by an orthogonal similarity transformation:
// Q**T * A * Q = T.
// N is the dimension of matrix
// A is NxN matrix, overwritten
// D is N   array; diagonal part
// E is N-1 array; superdiagonal part
// tau is
void gmp_dsytrdU(int N, mpf_class *A, int LDA, mpf_class *d, mpf_class *e, mpf_class *tau)
{
  mpf_class one;
  mpf_class mone;
  mpf_class zero;
  mpf_class half;
  mpf_class taui;
  mpf_class alpha;
  one = 1.0;
  mone = -1.0;
  zero = 0.0;
  half = 1.0/2.0;

  if(N<=0) return;
// Reduce the upper triangle of A
  for (int i=N-2; i>=0; i--){
//Generate elementary reflector H(i) = I - tau * v * v'
    //to annihilate A(1:i-1,i+1)
    gmp_dlarfg( i+1, &A[i+(i+1)*LDA], &A[0+(i+1)*LDA], 1, &taui);
    e[i] = A[i+(i+1)*LDA];
    if (taui!=zero){
      //Apply H(i) from both sides to A(1:i,1:i)
      A[i+(i+1)*LDA] = one;
      //Compute  x := tau * A * v  storing x in TAU(1:i)
      gmp_dsymvU(i+1, taui, A, N, &A[0+(i+1)*LDA], 1, zero, tau, 1);
      //Compute  w := x - 1/2 * tau * (x'*v) * v
      alpha = mone * half * taui * gmp_ddot( i+1, &tau[0], 1, &A[0+(i+1)*LDA], 1 );
      gmp_daxpy( i+1, alpha, &A[0+(i+1)*LDA], 1, tau, 1 );
      //Apply the transformation as a rank-2 update:
      //A := A - v * w' - w * v'
      gmp_dsyr2U( i+1, mone, &A[0+(i+1)*LDA], 1, &tau[0], 1, &A[0], N);
      A[i+(i+1)*LDA] = e[i];
    }
    d[i+1] = A[(i+1)+(i+1)*LDA];
    tau[i]=taui;
  }
  d[0] = A[0];
}

//http://www.netlib.org/lapack/double/dsterf.f
//DSTERF computes all eigenvalues of a symmetric tridiagonal matrix
//using the Pal-Walker-Kahan variant of the QL or QR algorithm.
int gmp_dsterf(int N, mpf_class *D, mpf_class *E)
{
  mpf_class zero, one, two, three;
  mpf_class sigma;
  mpf_class eps, eps2;
  mpf_class safmin, safmax, ssfmax, ssfmin, anorm;
  mpf_class rte, rt1, rt2, s, c, r, oldc, oldgam, gamma, p, bb, alpha;

  int maxit=30;
  int nmaxit;
  int info, iscale;
  int l1, jtot;
  int i, l, m, n;
  int lsv, lend, lendsv;

  zero = 0.0;
  one = 1.0;
  two = 2.0;
  three = 3.0;

  info = 0;

//Quick return if possible
  if(N<0)
    return -1;
  if(N<=1) return 0;

// Determine the unit roundoff for this environment.
  eps = gmp_dlamchE();
  eps2 = eps * eps;
  safmin = gmp_dlamchS();
  safmax = one / safmin;
  ssfmax = sqrt(safmax)/ three;
  ssfmin = sqrt(safmin)/ eps2;

// Compute the eigenvalues of the tridiagonal matrix.
  nmaxit = N*maxit;
  sigma = zero;
  jtot = 0;
  l1 = 0;
  lend = 0; l = 0; //dummy values at least lend==l
  while (jtot < nmaxit ) {
//Determine where the matrix splits and choose QL or QR iteration
//for each block, according to whether top or bottom diagonal
//element is smaller.
    if(l1>N-1) break;
    if(l1>0) E[l1-1]=zero;
    for (m=l1; m<N-1; m++){
      if( abs(E[m]) <=  sqrt(abs(D[m])) * sqrt(abs(D[m+1])) * eps) {
        E[m]=zero;
        break;
      }
    }
    l = l1;
    lend = m;
    l1 = m+1;
    lsv = l;  //save l and lend in lsv and lendsv
    lendsv = lend;
//XXX NOT IMPLEMENTED scale matrix
    anorm = gmp_dlanstI(lend-l+1, &D[l], &E[l]);
    iscale = 0;
    if (anorm > ssfmax) {
      iscale = 1;
      printf("# gmp_dsterf XXX scale by ssfmax not implemented \n");
    } else if (anorm < ssfmin) {
      iscale = 2;
      printf("# gmp_dsterf XXX scale by ssfmin not implemented \n");
    }
    for (i=l; i<lend; i++) {
      E[i] = E[i]*E[i];
    }
//Choose between QL and QR iteration
    if (abs(D[lend]) < abs(D[l])) {
      lend = lsv;
      l = lendsv;
    }
    if (lend>=l) {
//QL iteration
      while(1) {
//look for small subdiagonal element.
	if (l!=lend) {
	  for(m = l; m<lend; m++) {
            if (abs(E[m]) <= eps2*abs(D[m]*D[m+1]) ) break;
	  }
	} else { m = lend; }
        if (m<lend)
          E[m]=zero;
        p = D[l];
	if (m!=l) {
//if remaining matrix is 2 by 2, use dlae2 to compute its eigenvalues.
	  if (m==l+1) {
            rte = sqrt (E[l]);
            gmp_dlae2(D[l], rte, D[l+1], &rt1, &rt2);
            D[l] = rt1;
            D[l+1] = rt2;
            E[l] = zero;
            l = l + 2;
            if (l <=lend) continue;
            break; //goto 150
	  }
          if(jtot==nmaxit) break;
          jtot++;
//form shift
          rte = sqrt (E[l]);
          sigma = ( D[l+1] - p ) / (two*rte);
          r = gmp_dlapy2( sigma, one);
          sigma = p - (rte / (sigma + gmp_sign(r, sigma)));
          c = one;
          s = zero;
          gamma = D[m] - sigma;
          p = gamma * gamma;
//inner loop
	  for (i=m-1;i>=l;i--){
            bb = E[i];
            r = p + bb;
            if (i!=m-1)
              E[i+1] = s*r;
            oldc = c;
            c = p / r;
            s = bb / r;
            oldgam = gamma;
            alpha = D[i];
            gamma = c *(alpha-sigma) - s*oldgam;
            D[i+1] = oldgam + (alpha-gamma);
            if (c!=zero)
              p = (gamma*gamma) / c;
            else
              p = oldc * bb;
	  }
          E[l]=s*p;
          D[l]=sigma+gamma;
          continue;
	}
//Eigenvalue found.
        D[l] = p;
        l++;
        if ( l<=lend )
          continue;
        break;
      }
    } else {
//QR Iteration
      while(1) {
//look for small super diagonal element.
	if (l!=lend) {
	  for(m = l; m>=lend+1; m--) {
            if (abs(E[m-1]) <= eps2*abs(D[m]*D[m-1]) ) break;
	  }
	} else { m=lend; }
        if (m>lend)
          E[m-1]=zero;
        p = D[l];
	if (m!=l) {
//if remaining matrix is 2 by 2, use dlae2 to compute its eigenvalues.
	  if (m==l-1) {
            rte = sqrt (E[l-1]);
            gmp_dlae2(D[l], rte, D[l-1], &rt1, &rt2);
            D[l] = rt1;
            D[l-1] = rt2;
            E[l-1] = zero;
            l = l - 2;
            if (l>=lend) continue;
            break;
	  }
          if(jtot==nmaxit) break;
          jtot++;
//form shift
          rte = sqrt (E[l-1]);
          sigma = ( D[l-1] - p ) / (two*rte);
          r = gmp_dlapy2( sigma, one);
          sigma = p - (rte / (sigma + gmp_sign(r, sigma)));
          c = one;
          s = zero;
          gamma = D[m] - sigma;
          p = gamma * gamma;
//inner loop
	  for (i=m;i<=l-1;i++){
            bb = E[i];
            r = p + bb;
            if (i!=m)
              E[i-1] = s*r;
            oldc = c;
            c = p / r;
            s = bb / r;
            oldgam = gamma;
            alpha = D[i+1];
            gamma = c *(alpha-sigma) - s*oldgam;
            D[i] = oldgam + (alpha-gamma);
            if (c!=zero)
              p = (gamma*gamma) / c;
            else
              p = oldc * bb;
	  }
          E[l-1]=s*p;
          D[l]=sigma+gamma;
          continue;
	}
//Eigenvalue found.
        D[l] = p;
        l--;
        if ( l>=lend )
          continue;
        break;
      }
//XXX NOT IMPLEMENTED rescale matrix
      if (iscale == 1 ){
        printf("# gmp_dsterf XXX rescale by ssfmax not implemented \n");
      } else if (iscale == 2 ) {
        printf("# gmp_dsterf XXX rescale by ssfmin not implemented \n");
      }
    jtot++;
    }
  }
  if(jtot>nmaxit) {
    for (int i=0; i<N-1; i++) {
      if(E[i]!=zero) info++;
    }
  }
  qsort(D, N, sizeof(mpf_class), (int (*)(const void*, const void*))compare_mpf);
  return 0;
}

// http://www.netlib.org/lapack/double/dlaev2.f
// DLAEV2 computes the eigendecomposition of a 2-by-2 symmetric matrix
//    [  A   B  ]
//    [  B   C  ].
// On return, RT1 is the eigenvalue of larger absolute value, RT2 is the
// eigenvalue of smaller absolute value, and (CS1,SN1) is the unit right
// eigenvector for RT1, giving the decomposition
//    [ CS1  SN1 ] [  A   B  ] [ CS1 -SN1 ]  =  [ RT1  0  ]
//    [-SN1  CS1 ] [  B   C  ] [ SN1  CS1 ]     [  0  RT2 ].
void gmp_dlaev2(mpf_class a, mpf_class b, mpf_class c, mpf_class *rt1, mpf_class *rt2, mpf_class *cs1, mpf_class *sn1)
{
  mpf_class sm, df, adf, abs, tb, ab;
  mpf_class zero, one, two, half, acs;
  mpf_class acmx, acmn;
  mpf_class cs, ct, tn, rt;
  int sgn1, sgn2;

  zero = 0.0;
  one = 1.0;
  two = 2.0;
  half = 0.5;

//Compute the eigenvalues

  sm = a + c;
  df = a - c;
  adf = mpf_abs_(df);
  tb = b + b;
  ab = mpf_abs_(tb);
  if (mpf_abs_(a)>mpf_abs_(c)) {
    acmx = a;
    acmn = c;
  } else {
    acmx = c;
    acmn = a;
  }
  if (adf>ab) {
    rt = adf * sqrt ( one + (ab/adf) * (ab/adf) );
  } else if (adf<ab) {
    rt = ab * sqrt ( one + (adf/ab) * (adf/ab) );
  } else {
//Includes case AB=ADF=0
    printf("#gmp_dlaev2 Checkpoint 5 Not checked\n");
    rt = ab * sqrt(two);
  }
  if(sm<zero){
    *rt1 = half * (sm-rt);
    sgn1 = -1;
//Order of execution important.
//To get fully accurate smaller eigenvalue,
//next line needs to be executed in higher precision.
    *rt2 = (acmx / (*rt1))*acmn - (b/ (*rt1)) *b;
  } else if (sm>zero) {
    *rt1 = half * (sm+rt);
    sgn1 = 1;
//Order of execution important.
//To get fully accurate smaller eigenvalue,
//next line needs to be executed in higher precision.
    *rt2 = (acmx / (*rt1))*acmn - (b/(*rt1))*b;
  } else {
    printf("#gmp_dlaev2 Checkpoint 8 Not checked\n");
//Includes case RT1 = RT2 = 0
    *rt1 = half * rt;
    *rt2 = -1.0 * half * rt;
    sgn1 = 1;
  }
//Compute the eigenvector
  if (df>=zero) {
    cs = df + rt;
    sgn2 = 1;
  } else {
    cs = df - rt;
    sgn2 = -1;
  }
  acs = mpf_abs_(cs);
  if (acs>ab) {
    ct = -tb / cs;
    *sn1 = one / sqrt (one+ct*ct);
    *cs1 = ct* (*sn1);
  } else {
    printf("#gmp_dlaev2 Checkpoint 9 Not checked\n");
    if(ab==zero) {
      *cs1 = one;
      *sn1 = zero;
    } else {
    printf("#gmp_dlaev2 Checkpoint 13 Not checked\n");
      tn = -cs / tb;
      *cs1 = one /sqrt(one+tn*tn);
      *sn1 = tn* (*cs1);
    }
  }
  if(sgn1==sgn2) {
    tn=*cs1;
    *cs1 = -(*sn1);
    *sn1 = tn;
  }
  return;
}

//http://www.netlib.org/lapack/double/dlartg.f
// DLARTG generate a plane rotation so that
//    [  CS  SN  ]  .  [ F ]  =  [ R ]   where CS**2 + SN**2 = 1.
//    [ -SN  CS  ]     [ G ]     [ 0 ]
//
//  This is a slower, more accurate version of the BLAS1 routine DROTG,
//   *  with the following other differences:
//   F and G are unchanged on return.
//   If G=0, then CS=1 and SN=0.
//    If F=0 and (G .ne. 0), then CS=0 and SN=1 without doing any
//       floating point operations (saves work in DBDSQR when
//        there are zeros on the diagonal).
// If F exceeds G in magnitude, CS will be positive.
void gmp_dlartg(mpf_class f, mpf_class g, mpf_class *cs, mpf_class *sn, mpf_class *r)
{
  mpf_class zero;
  mpf_class one;
  mpf_class two;
  mpf_class f1, g1;
  int count;

  zero = 0.0;
  one = 0.0;
  two = 0.0;

  mpf_class safmin;
  mpf_class safmn2;
  mpf_class safmx2, eps, scale;

  safmin = gmp_dlamchS();
  eps = gmp_dlamchE();
// SAFMN2 = DLAMCH( 'B' )**INT( LOG( SAFMIN / EPS ) / LOG( DLAMCH( 'B' ) ) / TWO );
//        ~ 2^(ln(safmin/eps) / 2ln2 ) (dlamchB=2)  = sqrt(safmin/eps).
  safmn2 = sqrt(safmin / eps);
  safmx2 = 1.0 / safmn2;

  if (g==zero) {
    *cs = one;
    *sn = zero;
    *r = f;
  } else if (f==zero) {
    *cs = zero;
    *sn = one;
    *r = g;

  } else {
    f1 = f;
    g1 = g;
    scale = max (abs(f1), abs(g1));
    count = 0;
    if (scale >= safmx2 ){
      printf("#XXX gmp_dlartg not yet implemented #1 \n");
      while (1) {
        count++;
        f1 = f1 * safmn2;
        g1 = g1 * safmn2;
        scale = max (abs(f1), abs(g1));
        if (scale >= safmx2 ) continue;

        *r = sqrt ( f1*f1 + g1*g1);
        *cs = f1/(*r);
        *sn = g1/(*r);
	for (int i=0; i<count; i++){
          *r = (*r) * safmx2;
	}
        break;
      }
    } else if (scale<=safmn2) {
      printf("#XXX gmp_dlartg not yet implemented #3 \n");
      while(1){
        count++;
        f1 = f1 * safmx2;
        g1 = g1 * safmx2;
        scale = max (abs(f1), abs(g1));
        if (scale <= safmn2 ) continue;
        *r = sqrt (f1*f1+g1*g1);
        *cs = f1/(*r);
        *sn = g1/(*r);
	for (int i=0; i<count; i++){
          *r = (*r)*safmn2;
	}
        break;
      }
    } else {
      *r = sqrt (f1*f1+g1*g1);
      *cs = f1/(*r);
      *sn = g1/(*r);
    }
    if (abs(f)> abs(g) && (*cs)<zero) {
      *cs = -(*cs);
      *sn = -(*sn);
      *r = -(*r);
    }
  }
  return;
}

//http://www.netlib.org/lapack/double/dsteqr.f
// DSTEQR computes all eigenvalues and, optionally, eigenvectors of a
// symmetric tridiagonal matrix using the implicit QL or QR method.
// The eigenvectors of a full or band symmetric matrix can also be found
// if DSYTRD or DSPTRD or DSBTRD has been used to reduce this matrix to
// tridiagonal form.
// Trailing `I' denots that we compute both eigenvalues and eigenvectors.
int gmp_dsteqrI(int N, mpf_class *D, mpf_class *E, mpf_class *Z, int LDZ, mpf_class *work)
{
  int info, nmaxit, maxit, jtot, l1, nm1;
  int i, m, nm, mm, mm1, l, lm1, lend, lsv, lendsv, lendm1, iscale;

  mpf_class zero, one, two, three;
  mpf_class eps, eps2, safmin, safmax, ssfmax, ssfmin;
  mpf_class c, s, rt1, rt2;
  mpf_class tst, anorm, po;
  mpf_class f, b, p, g, r;
  maxit = 30;
  zero = 0.0;
  one = 1.0;
  two = 2.0;
  three = 3.0;

  info = 0;

  if (N==0) return info;
  if (N==1) {
    Z[0]=one; return info;
  }

//Determine the unit roundoff for this environment.
  eps = gmp_dlamchE();
  eps2 = eps * eps;
  safmin = gmp_dlamchS();
  safmax = one / safmin;
  ssfmax = sqrt(safmax)/ three;
  ssfmin = sqrt(safmin)/ eps2;

//Compute the eigenvalues and eigenvectors of the tridiagonal matrix.
  gmp_dlasetF(N, N, zero, one, Z, LDZ);
  nmaxit = N*maxit;
  jtot = 0;
//Determine where the matrix splits and choose QL or QR iteration
//for each block, according to whether top or bottom diagonal
//element is smaller.
  l1 = 0;
  nm1 = N-2;
//
  while (jtot < nmaxit ) {
    if(l1>N-1) break;
    if(l1>0) E[l1-1]=zero;
    if(l1<=nm1) {
      for (m=l1; m<N-1; m++){
        tst = abs(E[m]);
        if (tst==zero) break;
	if( tst <=  sqrt(abs(D[m])) * sqrt(abs(D[m+1])) * eps) {
          E[m]=zero;
          break;
	}
      }
    }
    l = l1;
    lend = m;
    l1 = m+1;
    lsv = l;
    lendsv = lend;
    if (lend == l ) continue;

//Scale submatrix in rows and columns L to LEND
//XXX not implemented yet.
    anorm = gmp_dlanstI(lend-l+1, &D[l], &E[l]);
    iscale = 0;
    if (anorm > ssfmax) {
      iscale = 1;
      printf("# gmp_dsteqr XXX ssfmax not implemented \n");
    } else if (anorm < ssfmin) {
      iscale = 2;
      printf("# gmp_dsteqr XXX ssfmin not implemented \n");
    }
//Choose between QL and QR iteration
    if (abs(D[lend]) < abs(D[l])) {
      lend = lsv;
      l = lendsv;
    }
    if (lend>=l) {

//QL iteration
      while(1) {
//look for small subdiagonal element.
	if (l!=lend) {
	  for(m = l; m<lend; m++) {
            tst = abs (E[m]*E[m]);
            if ( tst <= eps2*abs(D[m]*D[m+1])+safmin ) break;
	  }
	} else { m = lend; }
        if(m<lend) E[m]=zero;
        p = D[l];
	if (m!=l) {
//If remaining matrix is 2-by-2, use DLAEV2 to compute its eigensystem.
	  if (m==l+1) {
            gmp_dlaev2(D[l], E[l], D[l+1], &rt1, &rt2, &c, &s);
            work[l] = c;
            work[N-1+l] = s;
            gmp_dlasrRVB(N, 2, &work[l], &work[N-1+l], &Z[l*LDZ], LDZ);
            D[l]=rt1;
            D[l+1]=rt2;
            E[l]=zero;
            l = l+2;
            if(l<=lend) continue;
            break;
	  }
          if(jtot==nmaxit) continue;
          jtot++;
//form shift
          g = ( D[l+1]-p ) / ( two*E[l] );
          r = gmp_dlapy2 ( g, one );
          g = D[m] - p + ( E[l] / ( g+gmp_sign(r, g)));
          s = one;
          c = one;
          p = zero;
//Inner loop
          mm1 = m - 1;
	  for (i = mm1; i>=l; i--) {
            f = s * E[i];
            b = c * E[i];
            gmp_dlartg (g, f, &c, &s, &r);
            if (i!=m-1)
              E[i+1] = r;
            g = D[i+1] - p;
            r = ( D[i] - g) * s + two * c * b;
            p = s * r;
            D[i+1] = g+p;
            g = c * r - b;
//Eigenvectors are desired, thus save rotations.
            work[i] = c;
            work[N-1+i] = -s;
	  }
//Apply saved rotations;
          mm = m - l + 1;
          gmp_dlasrRVB(N, mm, &work[l], &work[N-1+l], &Z[l*LDZ], LDZ); //N should be LDZ...
          D[l] = D[l] - p;
          E[l] = g;
          continue;
	}
//Eigenvalue found
        D[l] = p;
        l++;
        if (l<=lend) continue;
        break;
      }
    } else {
//QR iteration
      while(1) {
	if (l!=lend) {
	  for(m = l; m>=lend+1; m--) {
            tst = abs (E[m-1]*E[m-1]);
	  if ( tst <= eps2*abs(D[m]*D[m-1])+safmin )  break;
	  }
	} else { m=lend; }

      if(m>lend) E[m-1]=zero;
        p = D[l];
	if (m!=l) {
//If remaining matrix is 2-by-2, use DLAEV2 to compute its eigensystem.
	  if (m==l-1) {
            gmp_dlaev2(D[l-1], E[l-1], D[l], &rt1, &rt2, &c, &s);
            work[m] = c;
            work[N-1+m] = s;
            gmp_dlasrRVF(N, 2, &work[m], &work[N-1+m], &Z[(l-1)*LDZ], LDZ); //N should be LDZ...
            D[l-1]=rt1;
            D[l]=rt2;
            E[l-1]=zero;
            l = l-2;
            if(l>=lend) continue;
            break;
	  }
          if(jtot==nmaxit) continue;
          jtot++;
//form shift
          g = ( D[l-1]-p ) / ( two*E[l-1] );
          r = gmp_dlapy2 ( g, one);
          g = D[m] - p + ( E[l-1] / ( g+gmp_sign(r, g)));
          s = one;
          c = one;
          p = zero;
//Inner loop
          lm1 = l - 1;
	  for (i = m; i<=lm1; i++) {
            f = s * E[i];
            b = c * E[i];
            gmp_dlartg (g, f, &c, &s, &r);
            if (i!=m)
              E[i-1] = r;
            g = D[i] - p;
            r = ( D[i+1] - g) * s + two * c * b;
            p = s * r;
            D[i] = g+p;
            g = c * r - b;
//Eigenvectors are desired, thus save rotations.
            work[i] = c;
            work[N-1+i] = s;
	  }
//Apply saved rotations;
          mm = l - m + 1;
          gmp_dlasrRVF(N, mm, &work[m], &work[N-1+m], &Z[m*LDZ], LDZ); //N should be LDZ...
          D[l] = D[l] - p;
          E[lm1] = g;
          continue;
	}
//Eigenvalue found
        D[l] = p;
        l--;
        if (l>=lend) continue;
        break;
      }

    }
//XXX NOT IMPLEMENTED rescale matrix
    if (iscale == 1 ){
      printf("gmp_dsteqrI XXX ssfmax not implemented \n");
    } else if (iscale == 2 ) {
      printf("gmp_dsteqrI XXX ssfmin not implemented \n");
    }
  }
  if(jtot>nmaxit) {
    for (int i=0; i<N-1; i++) {
      if(E[i]!=zero) info++;
    }
  }
//All eigenvalues are calculated
//Use selection sort to minize swaps of eigenvectors
  for (int ii = 1; ii<N; ii++) {
    int k, j;
    i = ii-1;
    k = i;
    p = D[i];
    for (j=ii; j<N; j++){
      if(D[j]<p) {
        k = j;
        p = D[j];
      }
    }
    if (k!=i){
      D[k]=D[i];
      D[i]=p;
      gmp_dswap(N, &Z[i*LDZ], 1, &Z[k*LDZ], 1);
    }
  }
  return info;
}

//http://www.netlib.org/lapack/double/dlaset.f
//DLASET initializes an m-by-n matrix A to BETA on the diagonal and
//ALPHA on the offdiagonals.
//F means full matrix.
void gmp_dlasetF(int M, int N, mpf_class alpha, mpf_class beta, mpf_class *A, int LDA)
{
  for(int j=0; j<N; j++){
    for(int i=0; i<M; i++){
      A[i+j*LDA] = alpha;
    }
  }

  for(int i=0; i<min(M,N); i++){
    A[i+i*LDA]=beta;
  }
  return;
}

//http://www.netlib.org/lapack/patch/src/dlasr.f
//see also : http://www.netlib.org/lapack/double/dlasr.f
//DLASR(RVB) performs the transformation
//A := A*P' ( Right-hand side; R) where P is
//P = P( 1 )*P( 2 )*...*P( z - 1 ) ( Backward sequence; B)
//where  P( k ) is a plane rotation matrix for the following planes:
//the plane ( k, k + 1 ) ( Variable pivot; V),
//c( k ) and s( k )  must contain the  cosine and sine that define the
//matrix  P( k ).  The two by two plane rotation part of the matrix
//P( k ), R( k ), is assumed to be of the form
//R( k ) = (  c( k )  s( k ) ).
//         ( -s( k )  c( k ) )
int gmp_dlasrRVB(int M, int N, mpf_class *C, mpf_class *S, mpf_class *A, int LDA)
{
  mpf_class ctmp;
  mpf_class stmp;
  mpf_class mtmp;
  mpf_class one  = 1.0;
  mpf_class zero = 0.0;
  for (int j=N-2; j>=0; j--){
    ctmp = C[j];
    stmp = S[j];
    if ( (ctmp!=one) || (stmp!=zero) ) {
      for (int i=0; i<M; i++) {
        mtmp = A[i+(j+1)*LDA];
        A[i+(j+1)*LDA] = ctmp * mtmp - stmp * A[i+j*LDA];
        A[i+j*LDA]     = stmp * mtmp + ctmp * A[i+j*LDA];
      }
    }
  }
  return 1;
}

//http://www.netlib.org/lapack/patch/src/dlasr.f
//see also : http://www.netlib.org/lapack/double/dlasr.f
int gmp_dlasrRVF(int M, int N, mpf_class *C, mpf_class *S, mpf_class *A, int LDA)
{
  mpf_class ctmp;
  mpf_class stmp;
  mpf_class mtmp;
  mpf_class one  = 1.0;
  mpf_class zero = 0.0;
  for (int j=0; j<N-1; j++){
    ctmp = C[j];
    stmp = S[j];
    if ( (ctmp!=one) || (stmp!=zero) ) {
      for (int i=0; i<M; i++) {
        mtmp = A[i+(j+1)*LDA];
        A[i+(j+1)*LDA] = ctmp * mtmp - stmp * A[i+j*LDA];
        A[i+j*LDA]     = stmp * mtmp + ctmp * A[i+j*LDA];
      }
    }
  }
  return 1;
}

//http://www.netlib.org/lapack/double/dlansy.f
//DLANSY  returns the value of the one norm,  or the Frobenius norm, or
//the  infinity norm,  or the  element of  largest absolute value  of a
//real symmetric matrix A.
mpf_class gmp_dlansyML(int N, mpf_class *A, int LDA)
{
  mpf_class value;
  value = 0.0;

  for(int i=0; i<N; i++){
    for(int j=0; j<=i; j++){
    value = mpf_abs_ (max (value, mpf_abs_(A[i+j*LDA])));
    }
  }
  return value;
}

//http://www.netlib.org/lapack/double/dsyev.f
//DSYEV computes all eigenvalues and, optionally, eigenvectors of a
//real symmetric matrix A. In this function, we do not calculate eigenvectors (N).
//and use lower triangular matrix (L)
//
int gmp_dsyevNL(int N, mpf_class *A, int LDA, mpf_class *W, mpf_class *work, int lwork)
{
  mpf_class zero, one;
  mpf_class safmin, eps, smlnum, bignum, rmin, rmax;
  mpf_class anrm, sigma;
  int i, iscale, inde, indtau, indwrk, llwork, lopt, info;
  int imax;

  zero = 0.0;
  one = 1.0;
  if (N==0){
    work[0]=one;
  }

  if(N==1){
    W[0] = A[0];
    work[0]=(double)3;
  }

  safmin=gmp_dlamchS();
  eps = gmp_dlamchP();
  smlnum = safmin / eps;
  bignum = one / smlnum;
  rmin = sqrt (smlnum);
  rmax = sqrt (bignum);

//Scale matrix to allowable range, if necessary
  anrm = gmp_dlansyML(N, A, LDA);
  iscale = 0;
  if (anrm > zero && anrm < rmin ) {
    iscale = 1;
    sigma = rmin / anrm;
  } else if (anrm > rmax ) {
    iscale = 1;
    sigma = rmax / anrm;
  }
  if (iscale == 1 ){
//    call dlascl( uplo, 0, 0, one, sigma, n, n, a, lda, info )
  printf("# gmp_dsyevNL: XXX scaling is not implemented yet\n");
  }
//Call DSYTRD to reduce symmetric matrix to tridiagonal form.
  inde = 0;
  indtau = inde + N;
  indwrk = indtau + N;
  llwork = lwork - indwrk + 1;

  gmp_dsytrdL(N, A, LDA, W, &work[inde], &work[indtau]);
  lopt = 2*N;

  info = gmp_dsterf(N, W, &work[inde]);

//If matrix was scaled, then rescale eigenvalues appropriately.
  if (iscale == 1 ) {
    if (info == 0 ) {
      imax = N;
    } else {
      imax = info - 1;
    }
//  gmp_dscal (imax, one / sigma, w, 1);
  printf("# gmp_dsyevNL: XXX scaling is not implemented yet\n");
  }

//Set WORK(0) to optimal workspace size.
  work[0] = lopt;
  return info;
}


//http://www.netlib.org/lapack/double/dsyev.f
//DSYEV computes all eigenvalues and, optionally, eigenvectors of a
//real symmetric matrix A. In this function, we calculate eigenvectors (V).
//and use lower triangular matrix (L)
//
int gmp_dsyevVL(int N, mpf_class *A, int LDA, mpf_class *W, mpf_class *work, int lwork)
{
#ifdef ZERO
  mpf_class zero, one;
  mpf_class safmin;
  int i;

  zero = 0.0;
  one = 1.0;

  if (N==0){
    work[0]=one;
  }

  if(N==1){
    W[0] = A[0];
    work[0]=(double)3;
    A[0] = one;
  }

  safmin=gmp_dlamchS();
  eps = gmp_dlamchP();
  smlnum = safmin / eps;
  bignum = one / smlmum;
  rmin = sqrt (smlnum);
  rmas = sqrt (bignum);

//Scale matrix to allowable range, if necessary
  anrm = gmp_dlansyML(N, A, LDA);
  iscale = 0;
  if (anrm > zero && anrm < rmin ) {
    iscale = 1;
    sigma = rmin / anrm;
  } else if (arnm > rmax ) {
    iscale = 1;
    sigma = rmax / anrm;
  }
  if (iscale == 1 ){
//    call dlascl( uplo, 0, 0, one, sigma, n, n, a, lda, info )
  printf("# gmp_dsyevVL: XXX scaling is not implemented yet\n");
  }
//Call DSYTRD to reduce symmetric matrix to tridiagonal form.
  inde = 0;
  indtau = inde + N;
  indwrk = indtau + N;
  llwork = lwork - indwrk + 1;

  gmp_dsytrdL(N, A, LDA, w, &work[inde], &work[indtau]);
  lopt = 2*N + work[indwrk];

//For eigenvectors, first call DORGTR to generate the orthogonal matrix, then call DSTEQR.
// CALL DORGTR( UPLO, N, A, LDA, WORK( INDTAU ), WORK( INDWRK ), LLWORK, IINFO )
  info = gmp_dsteqrI(N, w, &work[inde], A, LDA, &work[indtau]);

//If matrix was scaled, then rescale eigenvalues appropriately.
  if (iscale == 1 ) {
    if (info == 0 ) {
      imax = N;
    } else {
      imax = info - 1;
    }
//  gmp_dscal (imax, one / sigma, w, 1);
  printf("# gmp_dsyevVL: XXX scaling is not implemented yet\n");
  }
//Set WORK(0) to optimal workspace size.
  work[0] = lwkopt;

#endif
}


#ifdef __cplusplus
}
#endif

